package lsedit;

import java.io.*;



/**

 * A class to turn an TA input stream into a stream of tokens.

 * @version 1.0, 22/07/98

 * @author  Gary Farmaner

 */

public

class TA_StreamTokenizer extends Object {

    private InputStream is;

    private char buf[];

    private int peekc = ' ';

    private boolean pushedBack;

    private boolean forceLower;



    /** The line number of the last token read */

    private int LINENO = 1;



    private byte ctype[] = new byte[256];



    private static final byte CT_WHITESPACE = 1;

    private static final byte CT_DELIMIT    = 2;





    /**

     * The type of the last token returned.  It's value will either

     * be one of the following TT_* constants, or a single

     * character.  For example, if '+' is encountered and is

     * not a valid word character, ttype will be '+'

     */

    public int ttype;



    /**

     * The End-of-file token.

     */

    public static final int TT_EOF = -1;



    /**

     * The End-of-line token.

     */

    public static final int TT_EOL = '\n';



    /**

     * The number token.  This value is in nval.

     */

    public static final int TT_NUMBER = -2;



    /**

     * The word token.  This value is in sval.

     */

    public static final int TT_WORD = -3;



    /**

     * The Stream value.

     */

    public String sval;



    /**

     * The number value.

     */

    public double nval;



    public TA_StreamTokenizer (InputStream is) {

        this.is = is;

        buf = new char[20*1024];



        ctype[' ']  = (byte) (CT_WHITESPACE | CT_DELIMIT);

        ctype['\t'] = (byte) (CT_WHITESPACE | CT_DELIMIT);

        ctype['\r'] = (byte) (CT_WHITESPACE | CT_DELIMIT);

        ctype['\n'] = (byte) (CT_WHITESPACE | CT_DELIMIT);

	ctype['=']  = CT_DELIMIT;

	ctype[':']  = CT_DELIMIT;

	ctype['{']  = CT_DELIMIT;

	ctype['}']  = CT_DELIMIT;

	ctype['(']  = CT_DELIMIT;

	ctype[')']  = CT_DELIMIT;

    }



    protected void db() {

	System.out.println(LINENO + ": " + sval);

    }



    protected static final int IBUF_SIZE = 8 * 1024;



    protected byte[] ibuf = new byte[IBUF_SIZE];

    protected int cnt = 0;

    protected int pos = 0;



    protected final int cread() throws IOException {

	if (pos == cnt) {

	    cnt = is.read(ibuf, 0, IBUF_SIZE);

	    if (cnt < 1) {

		pos = cnt;

		return TT_EOF;

	    }

	    pos = 0;

	}



	return (int) ibuf[pos++];

    }



    protected final int creadNW(int curc) throws IOException {

	int c = curc;

	int pc = 0;



	while (true) {

	    switch(c)

	    {

	    case '\r':

                LINENO++;

		pc = '\r';

		break;



	    case '\n':

		if (pc != '\r')

		    LINENO++;

		// Fall through



	    case '\t':

	    case ' ':

		pc = c;

		break;



	    default:

		return (int) c;

	    }





	    if (pos < cnt)  {

		c = ibuf[pos++];

	    }

	    else {

		int nc = cread();



		if (nc == TT_EOF)

		    return TT_EOF;



		c = (byte) nc;

	    }

	}

    }



    protected final int creadNL() throws IOException {

	while(true) {

	    int c;



	    if (pos < cnt) {

		c = ibuf[pos++];

	    }

	    else {

		int nc = cread();



		if (nc == TT_EOF)

		    return TT_EOF;



		c = (byte) nc;

	    }



	    if (c == '\n') {

		LINENO++;

		return cread();

	    }

	    else if (c == '\r') {

		LINENO++;

		c = cread();

		if (c == '\n')

		    c = cread();

		return c;

	    }

	}

    }

    



    /**

     * Parses a token from the input stream.  The return value is

     * the same as the value of ttype.  Typical clients of this

     * class first set up the syntax tables and then sit in a loop

     * calling nextToken to parse successive tokens until TT_EOF

     * is returned.

     */

    public final int nextToken() throws IOException {

        if (pushedBack) {

            pushedBack = false;

            return ttype;

        }



        int c = peekc;

	int pc = 0;

	int pos = 0;



        sval = null;



        if (c == TT_EOF) {

	    sval = "EOF";

            return ttype = TT_EOF;

	}



	c = creadNW(c);



        if (c == TT_EOF) {

	    sval = "EOF";

            return ttype = TT_EOF;

	}





	switch(c)

	{

	case '/':

	    c = cread();

	    if (c == '/') {

	        // Toss the remainder of the line

		peekc = creadNL();

		return nextToken();

	    }

	    // Should handle C style comments too



	    if ((ctype[c] & CT_DELIMIT) != 0) {

		sval = "/";

		peekc = c;

	        // db();

		return ttype = TT_WORD;

	    }	



	    buf[pos++] = '/';

	    break;



	case '"':

	    do {

		c = cread();

		if (c != TT_EOF && c != '"') {

		    if (c == '\r' || (c == '\n' && pc != '\r')) {

			LINENO++;

		    }

		    buf[pos++] = (char) c;

		    pc = c;

		}

 	    } while (c != TT_EOF && c != '"');



	    peekc = cread();

	    sval = new String(buf, 0, pos);

	    // db();

	    return ttype = TT_WORD;

	}



	if (ctype[c] == CT_DELIMIT) {

	    sval = String.valueOf((char) c);

	    peekc = cread();

	    // db();

	    return ttype = c;

	}



	// Read characters until delimitter



	buf[pos++] = (char) c;



	while(true) {

	    c = cread();



	    if (c == TT_EOF || (ctype[c] & CT_DELIMIT) != 0) {

		peekc = c;

		sval = new String(buf, 0, pos);

	        // db();

		return ttype = TT_WORD;

	    }



	    buf[pos++] = (char) c;

	}

    }



    /** Does an unget of the last token */



    public final void pushBack() {

	pushedBack = true;

    } 



    /** Return the current line number. */

    public int lineno() {

        return LINENO;

    }

}

